////////////////////////////////////////////////////////////////////////
//  rpud : An R package for GPU computing
//  Copyright (C) 2010-2019 Chi Yau. All Rights Reserved.
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; version 3 of the License.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, see <http://www.gnu.org/licenses/>.


#include "rpud.h"
#include "inc/rpudev.h"


enum RPUDIST_TYPE {
    RPUDIST_UNKNOWN,
    RPUDIST_EUCLIDEAN,
    RPUDIST_MAXIMUM,
    RPUDIST_MANHATTAN,
    RPUDIST_CANBERRA,
    RPUDIST_BINARY,
    RPUDIST_MINKOWSKI
};


////////////////////////////////////////////////////////////////////////
// rpuMetricKernel
//
inline __device__ unsigned distoffset(
    const unsigned n, const unsigned x, const unsigned y) {

    return y+x*n-((x+1)*(x+2))/2;
}


namespace rpudist {

template <typename T, RPUDIST_TYPE t>
__global__ void rpuMetricKernel(
    cudaTextureObject_t texRef,
    T* d_output, unsigned num, unsigned dim, T p) {
        
    const int 
        xid = blockIdx.x * blockDim.x + threadIdx.x,
        yid = blockIdx.y * blockDim.y + threadIdx.y;
        
    // compute the distance between the x-th and y-th vectors
    if (xid < yid && yid < num) {

        unsigned num_item = 0;
        T dist = 0;
        for (unsigned i = 0; i < dim; i++) {
        
            const T 
                x = tex2D<T>(texRef, xid, i),
                y = tex2D<T>(texRef, yid, i);
            if (RPU_BOTH_NON_NA(x, y)) {
                switch (t) {
                
                case RPUDIST_EUCLIDEAN:
                {
                const T dev = x - y;
                if (RPU_NON_NA(dev)) {
                    dist += dev * dev;
                    num_item++;
                } }
                break;
                
                case RPUDIST_MAXIMUM:
                {
                const T dev = abs(x - y);
                if (RPU_NON_NA(dev)) {
                    if (dist < dev) dist = dev;
                    num_item++;
                } }
                break;
                
                case RPUDIST_MANHATTAN:
                {
                const T dev = abs(x - y);
                if (RPU_NON_NA(dev)) {
                    dist += dev;
                    num_item++;
                } }
                break;
                
                case RPUDIST_CANBERRA:
                {
                const T 
                    sum = abs(x) + abs(y),
                    diff = abs(x - y);
                if (sum > RPU_NUMERIC_MIN_DENORM || 
                    diff > RPU_NUMERIC_MIN_DENORM) {
                    T dev = diff/sum;
                    if (RPU_NON_NA(dev)) {
                        dist += dev;
                        num_item++;
                    } else if (!RPU_FINITE(diff) && diff == sum) {
                        dist += 1;
                        num_item++;
                    }
                } }
                break;
                
                case RPUDIST_BINARY:
                {
                if (RPU_BOTH_FINITE(x, y)) {
                    if (x || y) {
                        num_item++;
                        if (!(x && y)) dist++;
                    }
                } }
                break;
                
                case RPUDIST_MINKOWSKI:
                {
                const T dev = (x - y);
                if (RPU_NON_NA(dev)) {
                    dist += __powf(abs(dev), p);
                    num_item++;
                } }
                break;
                
                } // switch t
                
            } // if RPU

            
        } // for i
         
        if (num_item != dim) {
            dist *= T(dim)/T(num_item);
        }

        d_output[distoffset(num, xid, yid)] = 
            (t == RPUDIST_EUCLIDEAN) ? sqrt(dist) :
            (t == RPUDIST_MINKOWSKI) ? __powf(dist, T(1.0/p)) :
            dist;
            
    } // if xid

}


////////////////////////////////////////////////////////////////////////
// RPUDist
//
template <typename T>
class RPUDist {

public:
    explicit RPUDist(const char* type, const T* input,
        const unsigned num, const unsigned dim, const T power);

    virtual ~RPUDist();
     
    void find(T* output);

protected:
    const T* 
        m_input;
    const unsigned
        m_nx, m_ny, m_num,
        BLK_DX, BLK_DY;          // block dimensions
    const T
        m_power;

    RPUDIST_TYPE
        m_type;

    cudaArray_t
        m_cuArray;
    cudaTextureObject_t
        m_texRef;

    void initTexture();
};


template <typename T>
RPUDist<T>::RPUDist(const char* type, const T* input, 
        const unsigned num, const unsigned dim, const T power) :
    m_input(input), m_nx(num), m_ny(dim), m_num(num),
    m_power(power), BLK_DX(16), BLK_DY(16),
    m_cuArray(nullptr), m_texRef(0) {

    m_type =
        (!strcmp("euclidean", type))    ? RPUDIST_EUCLIDEAN :
        (!strcmp("maximum", type))      ? RPUDIST_MAXIMUM   :
        (!strcmp("manhattan", type))    ? RPUDIST_MANHATTAN :
        (!strcmp("canberra", type))     ? RPUDIST_CANBERRA  :
        (!strcmp("binary", type))       ? RPUDIST_BINARY    :
        (!strcmp("minkowski", type))    ? RPUDIST_MINKOWSKI :
        RPUDIST_UNKNOWN;
        
    if (m_type == RPUDIST_UNKNOWN) {
        throw rpu::AppException("Undefined distance type");
    }

    initTexture();
}


template <typename T>
RPUDist<T>::~RPUDist() {

    if (m_texRef) {
        cudaDestroyTextureObject(m_texRef);
        m_texRef = 0;
    }

    if (m_cuArray) {
        cudaFreeArray(m_cuArray);
        m_cuArray = nullptr;
    }
}


template <typename T>
void RPUDist<T>::initTexture() {

    ///////////////////////////////////////////////////////////////////
    // device input
    cudaChannelFormatDesc
        desc = cudaCreateChannelDesc<T>();
    rpuSafeCall(cudaMallocArray(
            &m_cuArray, &desc, m_nx, m_ny
        ));

    // copy input from host to device
    const unsigned
        INPUT_SIZE = sizeof(T)*(m_nx)*(m_ny);
    rpuSafeCall(cudaMemcpyToArray(
            m_cuArray, 0, 0, m_input, INPUT_SIZE,
            cudaMemcpyHostToDevice
        ));

    ///////////////////////////////////////////////////////////////////
    // create texture object
    cudaResourceDesc
        texRes;
    memset(&texRes, 0, sizeof(texRes));
    texRes.resType         = cudaResourceTypeArray;
    texRes.res.array.array = m_cuArray;

    cudaTextureDesc
        texDescr;
    memset(&texDescr, 0, sizeof(texDescr));
    texDescr.readMode = cudaReadModeElementType;

    rpuSafeCall(cudaCreateTextureObject(
            &m_texRef, &texRes, &texDescr, nullptr
        ));
}


template <typename T>
void RPUDist<T>::find(T* h_output) {

    RPU_MSG("RPUDist::find starts\n");
     
    const char *szError = nullptr;
    PFloat d_output = nullptr;
     
    try {
        ///////////////////////////////////////////////////////////////////
        // device output
        const unsigned OUTPUT_SIZE = sizeof(T)*(m_num)*(m_num-1)/2;
        rpuSafeCall(cudaMalloc((void**)&d_output, OUTPUT_SIZE));
    
        ///////////////////////////////////////////////////////////////////
        // kernel call
        const dim3 
            grid((m_num+BLK_DX-1)/BLK_DX, (m_num+BLK_DY-1)/BLK_DY), 
            blk(BLK_DX, BLK_DY);
                
        switch (m_type) {
        case RPUDIST_EUCLIDEAN:
            rpuMetricKernel<T, RPUDIST_EUCLIDEAN><<< grid, blk >>>(
                    m_texRef, d_output, m_nx, m_ny, m_power
                );
            break;
    
        case RPUDIST_MAXIMUM:
            rpuMetricKernel<T, RPUDIST_MAXIMUM><<< grid, blk >>>(
                    m_texRef, d_output, m_nx, m_ny, m_power
                );
            break;
    
        case RPUDIST_MANHATTAN:
            rpuMetricKernel<T, RPUDIST_MANHATTAN><<< grid, blk >>>(
                    m_texRef, d_output, m_nx, m_ny, m_power
                );
            break;
    
        case RPUDIST_CANBERRA:
            rpuMetricKernel<T, RPUDIST_CANBERRA><<< grid, blk >>>(
                    m_texRef, d_output, m_nx, m_ny, m_power
                );
            break;
    
        case RPUDIST_BINARY:
            rpuMetricKernel<T, RPUDIST_BINARY><<< grid, blk >>>(
                    m_texRef, d_output, m_nx, m_ny, m_power
                );
            break;
    
        case RPUDIST_MINKOWSKI:
            rpuMetricKernel<T, RPUDIST_MINKOWSKI><<< grid, blk >>>(
                    m_texRef, d_output, m_nx, m_ny, m_power
                );
            break;

        default:
            assert(false);
            break;
        }
        rpuCheckMsg("Kernel execution failed");

        // block until the device has completed
        rpuSafeCall(cudaDeviceSynchronize());

        // copy device output to host
        rpuSafeCall(cudaMemcpy(h_output, d_output, OUTPUT_SIZE, cudaMemcpyDeviceToHost));
    
        RPU_MSG("RPUDist::find ends\n");
     
    } catch (const std::exception& e) {
        szError = e.what();
    } catch (...) {
        szError = "Unknown error";
    }

    ///////////////////////////////////////////////////////////////////
    // free device memory
    if (d_output) cudaFree(d_output);
    
    ///////////////////////////////////////////////////////////////////
    // exit
    if (szError) error(szError);
    
}


}   // namespace rpudist


////////////////////////////////////////////////////////////////////////
// rpuDistance
//
void rpuDistance(
        const PString pType, 
        const PFloat points, 
        const PInteger pNum, 
        const PInteger pDim,
        const PFloat pMinkowski,
        PFloat pDistances) {

    String type = *pType;
    RPU_MSG("Finding %s distances ...\n", type);
     
    rpudist::RPUDist<Float> dist(
        type, 
        points, 
        static_cast<unsigned>(*pNum),
        static_cast<unsigned>(*pDim),
        *pMinkowski
        );
    dist.find(pDistances);
}






